require 'active_support/core_ext/object/blank'
require 'bob/compiler/message'
require 'bob/compiler/buffer'
require 'json'

module Bob
  module Compiler
    class Base
      def initialize
        @buffer = Buffer.new
        @warnings = []
        @errors = []
        @compiled = false
      end

      def warnings
        ensure_compiled { @warnings }
      end

      def errors
        ensure_compiled { @errors }
      end

      def warnings?
        ensure_compiled { @warnings.present? }
      end

      def well_formed?
        ensure_compiled { valid? && @warnings.blank? }
      end

      def errors?
        ensure_compiled { @errors.present? }
      end

      def valid?
        ensure_compiled { @errors.blank? }
      end

      def result
        ensure_compiled { @buffer.string }
      end

      alias_method :to_s, :result

      protected
        def buffer
          ensure_compiled { @buffer }
        end

      private
        def ensure_compiled
          compile! unless compiled?
          yield
        end

        def compiled?
          @compiled
        end

        def compile!
          @compiled = true
          compile
        end

        def compile
        end

        def warn(message, node = nil)
          @warnings << Message.new(message, node)
        end

        def error(message, node = nil)
          @errors << Message.new(message, node)
        end

        def identifier_from_attribute(node, attribute, required = true)
          sanitize = nil

          if node.has_attribute?(attribute)
            identifier = node[attribute]
            sanitized  = identifier && sanitize_identifier(identifier)

            if required && identifier.blank?
              error "Required attribute `#{attribute}` cannot be blank", node
            elsif required && sanitized.blank?
              error "Attribute `#{attribute}` does not contain a valid identifier", node
            elsif sanitized != identifier
              warn "Attribute `#{attribute}` does not contain a valid identifier, assuming you meant #{sanitized.inspect}", node
            end

          elsif required
            error "Missing required attribute `#{attribute}`", node
          end

          sanitized
        end

        def sanitize_identifier(identifier)
          sanitized = identifier.downcase

          # Replace all disallowed characters with a hyphen
          sanitized.gsub! /[^a-z0-9\-\/]+/, '-'

          # Each part must begin with a letter
          sanitized.gsub! /(?<=\A|\/)[^a-z]+/, ''

          # Remove any trailing hyphens
          sanitized.gsub! /-+(?=\z|\/)/, ''

          # Collapse multiple hyphens into one
          sanitized.gsub! /-+/, '-'

          # Collapse multiple slashes into one
          sanitized.gsub! /\/+/, '/'

          # Remove leading or trailing slashes
          sanitized.gsub! /\A\/|\/\z/, ''

          sanitized
        end

        def quote(arg)
          JSON.generate(arg, quirks_mode: true)
        end

        def pretty_quote(arg)
          JSON.pretty_generate(arg, quirks_mode: true)
        end
    end
  end
end
